// Unreal Fighting Framework by Kevin Poretti

#pragma once

#include "CoreMinimal.h"

/**
 *	CircleBuffer is a templated array data structure allowing for FIFO, O(1) insertion and removal,
 *	and wraparound when iterating/accessing elements of the buffer.
 *
 *	The underlying container type is a statically allocated array. You can use the second
 *	template parameter to specify the static size of the buffer. Defaults to 64.
 */
template<typename T, SIZE_T L = 64>
class UNREALFIGHTINGFRAMEWORK_API TCircleBuffer
{
public:
	TCircleBuffer()
	    : WriteIdx(0)
		, ReadIdx(0)
	{
	}

    /**
     * @brief Adds a new element to the buffer only if there is room.
     * @param Element to add to the buffer
     * @return true if there was room in the buffer and the element was added, false otherwise
     */
    bool Push(const T& Element)
    {
        if(_Num == L)
        {
            // buffer full so can't add element
            return false;
        }
        Buffer[WriteIdx] = Element;
        WriteIdx = (WriteIdx + 1) % L;
        _Num++;
        return true;
    }

    /**
     * @brief Adds a new element to the buffer.
     * This function always adds an element. If the buffer is full the oldest element will be overwritten.
     * @param Element to add to the buffer
     */
    void ForcePush(const T& Element)
	{
	    if(_Num == L)
        {
            // buffer is full so we need to overwrite the oldest element
            // and make the read index point to the next oldest element
            ReadIdx = (ReadIdx + 1) % L;
        }
        Buffer[WriteIdx] = Element;
        WriteIdx = (WriteIdx + 1) % L;
        _Num = FMath::Min(_Num + 1, L);
	}

    /**
     * @brief Returns the oldest element in the buffer and removes that element from the buffer.
     * @param Result out parameter to be filled in with the popped element
     * @return true if there was an element in the buffer to pop, false otherwise
     */
    bool Pop(T* Result)
    {
        if(IsEmpty())
        {
            return false;
        }

        *Result = Buffer[ReadIdx];
        ReadIdx = (ReadIdx + 1) % L;
        _Num--;
        return true;
    }

    /**
     * @brief Returns the newest element that was added to the buffer
     * @param Result out parameter to be filled in with the most recently added element
     * @return true if there was an element in the buffer to peak, false otherwise
     */
    bool PeakLast(T* Result)
    {
	    if(IsEmpty())
        {
            return false;
        }

        *Result = Buffer[((SSIZE_T)WriteIdx - 1) <= 0 ? L - 1 : (WriteIdx - 1)];
        return true;
    }

    /**
     * @brief Returns the oldest element that was added to the buffer
     * @param Result out parameter to be filled in with the oldest element
     * @return true if there was an element in the buffer to peak, false otherwise
     */
    bool PeakFirst(T* Result)
    {
	    if(IsEmpty())
        {
            return false;
        }

        *Result = Buffer[ReadIdx];
        return true;
    }

    /**
     * @brief Returns the element at an index supplied to the function.
     * This function will account for write index offset and wraparound of the supplied index.
     * i.e. Index 0 will be the oldest added element while index Num - 1 will be the newest added element
     * @param Index of the element to return
     * @return the element at the index supplied
     */
    T& operator[](SIZE_T Index)
    {
        return Buffer[((SSIZE_T)WriteIdx - (SSIZE_T)Index) < 0 ? L - (Index - WriteIdx) : WriteIdx - Index];
    }

    FORCEINLINE bool IsEmpty() const
	{
	    return _Num == 0;
	}

    /**
     * @brief Returns the max size of the underlying buffer
     * @return Max size of the underlying buffer
     */
    FORCEINLINE SIZE_T Max() const
	{
	    return L;
	}

    /**
     * @brief Returns the current number of elements of the underlying buffer
     * @return Max size of the underlying buffer
     */
    FORCEINLINE SIZE_T Num() const
	{
	    return _Num;
	}

private:
    T Buffer[L];
	SIZE_T WriteIdx;
	SIZE_T ReadIdx;
    SIZE_T _Num;
};